package czerwony.kapturek.jbwheel;


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;

import com.borland.jbuilder.JBuilderMenu;
import com.borland.jbuilder.JBuilder;
import com.borland.primetime.ide.*;
import com.borland.primetime.editor.*;
import com.borland.primetime.node.*;
import com.borland.primetime.actions.*;
import com.borland.primetime.properties.PropertyManager;
import com.borland.primetime.vfs.Url;

/** Wheel tool class */
public class WheelTool
{
    /* notch move */
    private final static int NOTCH_MOVE = 120;

    /** configuration structure */
    private Config conf = new Config();

    /** "not used" amount of wheel movement */
    private int reminder;

    /**
     * UGLY: if we cannot find anything suitable
     * for scrolling but the presence of a scrollbar is found
     * somewhow it can be used for scrolling. This is done mostly
     * for scrolling help windows (witch for some reason do
     * not have Scrollable components naither they use JScrollPane
     * or JViewport).
     * The only nice thing about it is that it can be used for
     * switching the orientation of scrolling from default (VERTICAL)
     * to HORIZONTAL is the mouse pointer is over the horizontal
     * JScrollBar (in a similar way some MS programs work on Windows).
     */
    private JScrollBar scrollFound;

    /**
     * On JDK 1.3 trying to scroll in a context of JNI call for some reason
     *  (I don't have time to invastigate) freezes the whole thing.
     *  Because of that scrolling is delegated to additional object invoked
     *  using SwingUtilities.invokeLater() method.
     */
    private WheelNotifier notifier;

    /**  */
    private final static String DLL_NAME = "jbWheel";

    /** The one and only instance of the tool */
    private static WheelTool I;

    private int action;

    /**
     * Method initOpenTool <TODO!!!>
     *
     *
     * @param major
     * @param minor
     */
    public static void initOpenTool(byte major, byte minor)
    {
        I = new WheelTool();
    }


    /**
     * Constructor WheelTool <TODO!!!>
     *
     */
    public WheelTool()
    {
        printWelcomeMessage();

        try
        {
            if (System.getProperty("java.vm.version", "").indexOf("1.2") == -1)
            {
                notifier = new WheelNotifier();
                trace(" ...running on JVM 1.3+");
            }

            initLibrary();
            initMenu();
        }
        catch (Exception ex)
        {
            //JOptionPane.showMessageDialog(null, ex.toString());
            ex.printStackTrace();
        }
    }


    private void printWelcomeMessage()
    {
        System.out.println(
            "**********************************************************************\n" +
            "**  Initializing: jbWheel                                           **\n" +
            "**                by Czerwony Kapturek (czerwony.kapturek@wp.pl)    **\n" +
            "**********************************************************************"
        );
    }

    /** Add to JBuilder menu action for configuring jbWheel tool. */
    private void initMenu()
    {
        Icon ico = new ImageIcon(getClass().getResource("mouse.gif"));

        JBuilderMenu.GROUP_ToolsOptions.add(new AbstractAction("WM", ico)
        {
            public void actionPerformed(ActionEvent ev)
            {
                JDialog dlg = new WheelToolDialog(Browser.getActiveBrowser(), conf);
                Rectangle r = Browser.getActiveBrowser().getBounds();
                Dimension d = dlg.getSize();
                r.x = (r.width - d.width) / 2;
                r.y = (r.height - d.height) / 2;
                r.width = d.width;
                r.height = d.height;

                dlg.setBounds(r);

                boolean debug = conf.debug;
                dlg.setModal(true);
                dlg.show();
                if (debug != conf.debug)
                    setDebug(conf.debug);
            }

            public String toString()
            {
                return "WheelMouse settings...";
            }
        });
    }

    /**
     * Load the native part and start capturing mousewheel messages.
     */
    private void initLibrary() throws Exception
    {
        String rootDir = JBuilder.getRootDir();
        String lib = System.mapLibraryName(DLL_NAME);
        Url url = PropertyManager.getInstallRootUrl().getRelativeUrl("lib/ext");
        File libFile = new File( url.getFileObject(), lib );
        if (libFile.exists())
        {
            trace("loading jbWheel.dll form: " + libFile.getCanonicalPath());
            System.load(libFile.getCanonicalPath());
        }
        else
        {
            trace("loading jbWheel.dll form java.lib.path");
            System.loadLibrary(DLL_NAME);
        }

        setDebug(conf.debug);
        startCaptureWheelMessages();
    }


    /** Sets the hook for intercepting wheel signals. */
    private native void startCaptureWheelMessages();

    /** Removes the hook for intercepting wheel signals. */
    private native void stopCaptureWheelMessages();

    /** Starts/stops printing debugging messagess in the DLL. */
    private native void setDebug(boolean d);


    /**
     * Entry point for the notifications about wheel moves.
     * Calls the routine responsible for processing notifications directly
     * (on 1.2)or indirectly (on 1.3).
     *
     * @param move
     * @param x
     * @param y
     * @param mask
     */
    public synchronized void onWheelMoved(int move, int x, int y, int mask)
    {
        trace("wheel move notification");

        try
        {
            if (notifier == null)
            {
                wheelMoved(move, x, y, mask);
            }
            else
            {
                notifier.move += move;
                notifier.x = x;
                notifier.y = y;
                notifier.mask = mask;

                SwingUtilities.invokeLater(notifier);
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }


    /**
     * Method wheelMoved <TODO!!!>
     *
     *
     * @param move
     * @param x
     * @param y
     * @param mask
     */
    public void wheelMoved(int move, int x, int y, int mask)
    {
        if (move * reminder >= 0)
        {
            move += reminder;
        }

        reminder = (conf.sensitivity * move) % (NOTCH_MOVE * 10);

        trace_i(" ...amount to scroll: ", move);

        if (move == 0)
        {
            return;
        }

        Integer i = (Integer)conf.actions.get(new Integer(mask));
        if (i == null)
        {
            return;
        }
        action = ((i.intValue() & Config.Action.REVERSE) != 0) ? i.intValue() - Config.Action.REVERSE : i.intValue();

        if ((i.intValue() & Config.Action.REVERSE) != 0)
        {
            move = -move;
            trace("reverse");
        }

        switch (action)
        {

            case Config.Action.nextnode :
                try
                {
                    if (move > 0)
                    {
                        Browser.getActiveBrowser().nextOpenNode(true);
                    }
                    else
                    {
                        Browser.getActiveBrowser().previousOpenNode(true);
                    }
                }
                catch (Exception ex)
                {
                }
                break;

            default :
                actionScroll(move, x, y);
                break;
        }
    }


    /**
     * Method actionScroll <TODO!!!>
     *
     *
     * @param move
     * @param x
     * @param y
     */
    private void actionScroll(int move, int x, int y)
    {
        scrollFound = null;

        JComponent c = null;

        if (conf.mode == Config.MODE_EDITORONLY)
        {
            c = findEditorPane();
        }
        else if (conf.mode == Config.MODE_UNDERCURSOR)
        {
            c = findComponentAt(x, y);
        }
        else if (conf.mode == Config.MODE_FOCUSED)
        {
            c = findFocused();
        }

        if (c instanceof Scrollable)
        {
            scrollScrollable(c, move);
        }
        else if (c instanceof JScrollPane)
        {
            scrollPanel((JScrollPane)c, move);
        }
        else if (c instanceof JSlider)
        {
            scrollSlider((JSlider)c, move);
        }
        else if (c instanceof JComboBox)
        {
            scrollCombo((JComboBox)c, move);
        }
        else if (scrollFound != null)
        {
            scrollScrollBar(move);
        }
        else
        {
            trace_o("nothing suitable to scroll: ", ((c == null) ? "null"
                                                                 : c.getClass()
                                                                     .getName()));
        }
    }


    /**
     * Method findEditorPane <TODO!!!>
     *
     *
     * @return
     */
    private EditorPane findEditorPane()
    {
        Browser b = Browser.getActiveBrowser();
        Node n = b.getActiveNode();
        NodeViewer v = b.getActiveViewer(n);
        JComponent c = v.getViewerComponent();
        return findEditorPane(c);
    }


    /**
     * Method findEditorPane <TODO!!!>
     *
     *
     * @param c
     *
     * @return
     */
    private EditorPane findEditorPane(Component c)
    {
        Component[] cc = ((JComponent)c).getComponents();
        for (int i = 0; i < cc.length; i++)
        {
            if (cc[i] instanceof EditorPane)
            {
                return (EditorPane)cc[i];
            }

            EditorPane ep = findEditorPane(cc[i]);
            if (ep != null)
            {
                return ep;
            }
        }

        return null;
    }


    /**
     * Method for searching for the focus owner.
     *
     * @return focused component.
     */
    private JComponent findFocused()
    {
        Window w = findTopWindow();
        if (w == null)
        {
            return null;
        }

        JComponent c = null;
        Component c1 = w.getFocusOwner();
        while (true)
        {
            if ((c1 instanceof Scrollable) || (c1 instanceof JScrollPane))
            {
                c = (JComponent)c1;

                break;
            }
            else if (c1 == null)
            {
                break;
            }

            c1 = c1.getParent();
        }

        return c;
    }


    /**
     * Method findComponentAt <TODO!!!>
     *
     *
     * @param x
     * @param y
     *
     * @return
     */
    private JComponent findComponentAt(int x, int y)
    {
        Point p = new Point(x, y);
        Component c = null;
        try
        {
            // first: find the window
            Container cp = findTopWindow();
            trace_o("top window: ", (cp == null) ? "null"
                                                 : cp.getClass().getName());

            if (cp == null)
            {
                return null;
            }

            SwingUtilities.convertPointFromScreen(p, cp);

            Component c1 = SwingUtilities.getDeepestComponentAt(cp, p.x, p.y);
            if (c1 instanceof JScrollBar)
            {
                scrollFound = (JScrollBar)c1;
            }

            // go up and return when something suitable for
            // scrolling found
            while (true)
            {
                trace_o("going up: ", ((c1 != null) ? c1.getClass().getName()
                                                    : "NULL"));

                if ((c1 instanceof Scrollable) || (c1 instanceof JScrollPane)
                        || (c1 instanceof JSlider) || (c1 instanceof JComboBox))
                {
                    // System.err.println( "returning " + c1.getClass().getName() );
                    c = (JComponent)c1;

                    break;
                }
                else if (c1 == null)
                {
                    break;
                }
                else if (c1 instanceof Container)
                {
                    Component[] allComp = ((Container)c1).getComponents();
                    JScrollBar bar = null;

                    for (int n = 0; n < allComp.length; n++)
                    {
                        if (allComp[n] instanceof JScrollBar)
                        {
                            bar = (JScrollBar)allComp[n];

                            int orientation =
                                ((action == Config.Action
                                .hscroll) || (action == Config.Action
                                    .pghscroll)) ? JScrollBar.HORIZONTAL
                                                 : JScrollBar.VERTICAL;
                            if (bar.getOrientation() == orientation)
                            {
                                scrollFound = bar;
                                trace("JScrollbar (brother) - orientation OK");
                                return bar;
                            }
                        }
                    }

                    if (bar != null)
                    {
                        scrollFound = bar;
                        trace("JScrollbar (brother) - wrong orientation");
                        return bar;
                    }
                }

                c1 = c1.getParent();
            }

            return (JComponent)c;

        }
        catch (ClassCastException ex)
        {
            ex.printStackTrace();

            return null;
        }
    }



    /**
     * ugly but works
     *
     * @return
     */
    private Window findTopWindow()
    {
        Browser b = Browser.getActiveBrowser();
        Component c = null;
        Frame[] fAll = b.getFrames();
        for (int i = 0; i < fAll.length; i++)
        {
            trace_o(fAll[i].isVisible() ? "application frame (visible): "
                                        : "application frame (hidden): ", fAll[i]
                                            .getTitle());

            Window w = findTopWindow(fAll[i], 1);
            if (w != null)
            {
                return w;
            }

            c = fAll[i].getFocusOwner();

            if (c != null)
            {
                return fAll[i];
            }
        }

        return null;
    }


    /**
     * Method findTopWindow <TODO!!!>
     *
     *
     * @param w
     * @param level
     *
     * @return
     */
    private Window findTopWindow(Window w, int level)
    {
        Component c;
        Window[] wAll = w.getOwnedWindows();
        for (int j = 0; j < wAll.length; j++)
        {
            if (conf.debug)
            {
                trace("window (" + level + "): " + wAll[j].getName()
                      + (wAll[j].isVisible() ? "" : "(hidden)"));
            }

            // it looks like JBuilder 3.5 has a bug: even if dialog is
            // closed it's still on "owned windows" list.
            if (wAll[j].isVisible())
            {
                c = wAll[j].getFocusOwner();

                if (c != null)
                {
                    return wAll[j];
                }
            }

            Window w1 = findTopWindow(wAll[j], level++);
            if (w1 != null)
            {
                return w1;
            }
        }

        return null;
    }



    /**
     * Method scrollScrollable <TODO!!!>
     *
     * @param comp
     * @param move
     */
    private void scrollScrollable(JComponent comp, int move)
    {
        trace("scrollScrollable()");

        int orient = ((scrollFound != null) && conf.followBar)
                     ? scrollFound.getOrientation()
                     : (((action == Config.Action.vscroll) || (action == Config.Action.pgvscroll))
                        ? SwingConstants.VERTICAL
                        : SwingConstants.HORIZONTAL);

        Rectangle r = calculateScrolledRect(comp, move, orient);
        if ((r == null) && conf.bothDirections
            && (calculateScrolledRect(comp, -move, orient) == null))
        {
            orient = (orient == SwingConstants.VERTICAL)
                        ? SwingConstants.HORIZONTAL
                        : SwingConstants.VERTICAL;
            r = calculateScrolledRect(comp, move, orient);
        }
        if (r !=null )
        {
            comp.scrollRectToVisible(r);
        }
    }


    private Rectangle calculateScrolledRect(JComponent comp, int move, int orient)
    {
        Rectangle r = new Rectangle();
        comp.computeVisibleRect(r);

        int i = 0;
        if ((action == Config.Action.pgvscroll) || (action == Config.Action.pghscroll))
        {
            i = ((Scrollable)comp).getScrollableBlockIncrement(r, orient, -move);
            if (move < 0)
            {
                i = -i;
            }
        }
        else
        {
            int incr = ((Scrollable)comp).getScrollableUnitIncrement(r, orient, -move);
            i = (conf.sensitivity * move / (NOTCH_MOVE * 10)) * incr;
        }

        if (orient == SwingConstants.VERTICAL)
        {
            int ry = r.y;
            r.y = calculateNewOrigin(r.y, r.height, comp.getHeight(), i);
            if (ry == r.y)
                return null;
        }
        else
        {
            int rx = r.x;
            r.x = calculateNewOrigin(r.x, r.width, comp.getWidth(), i);
            if (rx == r.x)
                return null;
        }

        return r;
    }


    private int calculateNewOrigin(int oldOrig, int rectDim, int compDim, int delta)
    {
        oldOrig -= delta;

        if (oldOrig < 0)
        {
            oldOrig = 0;
        }
        else
        {
            oldOrig = Math.min(compDim - rectDim, oldOrig);
        }

        return oldOrig;
    }


    /**
     * JScrollPanel scrolling function
     *
     * @param comp
     * @param move
     */
    private void scrollPanel(JScrollPane comp, int move)
    {
        trace("scrollPanel()");
        JViewport vp = comp.getViewport();
        if (vp.getView() instanceof Scrollable)
        {   //just in case: shouldn't happen
            scrollScrollable((JComponent)vp.getView(), move);
            return;
        }

        JScrollBar sb = ((scrollFound != null) && conf.followBar)
            ? scrollFound
            : (((action == Config.Action.vscroll) || (action == Config.Action.pgvscroll))
                ? comp.getVerticalScrollBar()
                : comp.getHorizontalScrollBar());

        if (sb != null)
        {
            if (!sb.isVisible() && conf.bothDirections)
            {
                JScrollBar sb1 = sb.getOrientation()==JScrollBar.HORIZONTAL
                        ? comp.getVerticalScrollBar()
                        : comp.getHorizontalScrollBar();
                if (sb1 != null)
                {
                    sb = sb1;
                }
            }

            scrollFound = sb;
            trace(sb.getOrientation()==sb.HORIZONTAL ? "H" : "V");
            scrollScrollBar(move);
        }
    }



    /**
     * Helper function for scroll panels.
     * Used also if there is nothing better to scroll. Needed for Help
     * windows.
     *
     * @param move
     */
    private void scrollScrollBar(int move)
    {
        trace("scrollScrollBar()");
        int i = 0;
        if ((action == Config.Action.pgvscroll)
                || (action == Config.Action.pghscroll))
        {
            i = scrollFound.getBlockIncrement((move > 0) ? -1 : 1);
            if (move < 0)
            {
                i = -i;
            }
        }
        else
        {
            i = scrollFound.getUnitIncrement((move > 0) ? -1 : 1);
            i *= (conf.sensitivity * move / (NOTCH_MOVE * 5));
        }

        i = scrollFound.getValue() - i;
        i = Math.min(i, scrollFound.getMaximum());
        i = Math.max(i, scrollFound.getMinimum());
        scrollFound.setValue(i);
    }


    /**
     * Scrolling JSlider
     *
     * @param comp component to scroll
     * @param move unused
     */
    private void scrollSlider(JSlider comp, int move)
    {
        trace("scrollSlider()");
        int dv =
            ((action == Config.Action.hscroll) || (action == Config.Action
            .vscroll)) ? comp.getMinorTickSpacing()
                       : comp.getMajorTickSpacing();
        if (dv == 0) //no ticks
        {
            int d = comp.getMaximum() - comp.getMinimum();
            dv =
                ((action == Config.Action.hscroll) || (action == Config.Action
                .vscroll))
                    ? Math.max(d / 10, 1)
                    : Math.max(d / 4, 1);
        }
        if (move < 0)
        {
            dv = -dv;
        }

        int v = comp.getValue() + dv;
        v = Math.min(v, comp.getMaximum());
        v = Math.max(v, comp.getMinimum());

        comp.setValue(v);
    }


    /**
     * Method scrollCombo <TODO!!!>
     *
     *
     * @param comp
     * @param move
     */
    private void scrollCombo(JComboBox comp, int move)
    {
        trace("scrollCombo()");
        int idx = comp.getSelectedIndex() + ((move < 0) ? 1
                                                        : -1);
        idx = Math.min(idx, comp.getItemCount() - 1);
        idx = Math.max(idx, 0);

        comp.setSelectedIndex(idx);
    }


    /**
     * finalization - not used
     *
     * @throws Throwable
     */
    protected void finalize()
        throws Throwable
    {
        stopCaptureWheelMessages();
        trace("finalize() called - this cannot be true");
    }


    private final static MessageCategory WHEEL_MESSAGESS =
        new MessageCategory("wheel");

    /**
     * @param msg
     */
    private void trace(String msg)
    {
        if (conf.debug)
        {
            System.out.println("jbWheel: " + msg);
        }
    }


    /**
     * Tracing with single int parameter.
     *
     * @param msg
     * @param i
     */
    private void trace_i(String msg, int i)
    {
        if (conf.debug)
        {
            System.out.println("jbWheel: " + msg + i);
        }
    }


    /**
     * Tracing with single object parameter.
     *
     * @param msg
     * @param o
     */
    private void trace_o(String msg, Object o)
    {
        if (conf.debug)
        {
            System.out.println("jbWheel: " + msg + (o==null ? "null" : o.toString()));
        }
    }


    /**
     * An attempt to solve problem with freezing on JDK 1.3 & JB3.5.
     */
    private class WheelNotifier implements Runnable
    {
        int move;
        int x;
        int y;
        int mask;
        long time;

        public void run()
        {
            wheelMoved(move, x, y, mask);
            move = 0;
        }
    }

}


